import datetime
import re
from utils import unicode_or_None

from storm.locals import Date, Int, Reference, ReferenceSet, Unicode

ROADMAP_STATUSES_MAP = {
    u'Completed': [u'Implemented'],
    u'Blocked': [u'Needs Infrastructure', u'Blocked', u'Deferred'],
    u'In Progress': [u'Deployment', u'Needs Code Review',
                     u'Beta Available', u'Good progress',
                     u'Slow progress', u'Started'],
    u'Planned': [u'Unknown', u'Not started', u'Informational']}

ROADMAP_ORDERED_STATUSES = ['Completed', 'In Progress', 'Blocked', 'Planned']
assert set(ROADMAP_ORDERED_STATUSES) == set(ROADMAP_STATUSES_MAP.keys()), (
    'The roadmap statuses are incorrect: %s' % ROADMAP_ORDERED_STATUSES)


def fill_blueprint_info_from_launchpad(model_bp, lp_bp):
    model_bp.name = unicode_or_None(lp_bp.name)
    model_bp.url = unicode_or_None(lp_bp.web_link)
    model_bp.priority = unicode_or_None(lp_bp.priority)
    model_bp.implementation = unicode_or_None(
        lp_bp.implementation_status)
    if lp_bp.assignee_link is not None:
        assignee_name = extract_user_name_from_url(lp_bp.assignee_link)
        model_bp.assignee_name = unicode_or_None(assignee_name)
    model_bp.status = get_whiteboard_section(
        lp_bp.whiteboard, 'Status')
    if model_bp.status is None:
        model_bp.status = unicode_or_None(lp_bp.summary)
        if model_bp.status is None:
            model_bp.status = u"(no description)"
    if lp_bp.milestone_link is not None:
        milestone_name = extract_last_path_segment_from_url(
            lp_bp.milestone_link)
        model_bp.milestone_name = unicode_or_None(milestone_name)
    model_bp.details_url = unicode_or_None(lp_bp.specification_url)


def get_whiteboard_section(whiteboard, heading):
    if whiteboard is None:
        return None
    regex = '^' + heading + ':\s*\n(.*?)\n\n'
    m = re.search(regex, whiteboard, re.S | re.I | re.M)
    if m is not None:
        section = m.group(1).strip()
    else:
        section = None
    return section


class Milestone(object):

    __storm_table__ = "milestones"

    name = Unicode(primary=True)
    due_date = Date()
    project = Unicode()


def get_roadmap_status_for_bp_implementation_status(implementation):
    for key in ROADMAP_STATUSES_MAP:
        if implementation in ROADMAP_STATUSES_MAP[key]:
            return key
    # XXX: Is None the appropriate return value here?
    return None


class Blueprint(object):

    __storm_table__ = "specs"

    name = Unicode(primary=True)
    url = Unicode()
    priority = Unicode()
    implementation = Unicode()
    assignee_name = Unicode(name="assignee")
    team = Unicode()
    status = Unicode()
    milestone_name = Unicode(name="milestone")
    definition = Unicode()
    drafter_name = Unicode(name="drafter")
    approver_name = Unicode(name="approver")
    details_url = Unicode()
    roadmap_notes = Unicode()

    @classmethod
    def from_launchpad(cls, lp_bp):
        model_bp = cls()
        fill_blueprint_info_from_launchpad(model_bp, lp_bp)
        model_bp.definition = lp_bp.definition_status
        if lp_bp.drafter_link is not None:
            drafter_name = extract_user_name_from_url(lp_bp.drafter_link)
            model_bp.drafter_name = drafter_name
        if lp_bp.approver_link is not None:
            approver_name = extract_user_name_from_url(lp_bp.approver_link)
            model_bp.approver_name = approver_name
        model_bp.roadmap_notes = get_whiteboard_section(
            lp_bp.whiteboard, "Roadmap\s+Notes")
        return model_bp

    @property
    def roadmap_status(self):
        return get_roadmap_status_for_bp_implementation_status(
            self.implementation)


def current_date():
    return datetime.date.today()


class Person(object):

    __storm_table__ = "real_names"

    name = Unicode(primary=True)
    display_name = Unicode()


class TeamParticipation(object):

    __storm_table__ = "teams"
    __storm_primary__ = 'person_name', 'team_name'

    person_name = Unicode(name="name")
    team_name = Unicode(name="team")


class TeamStructure(object):

    __storm_table__ = "team_structure"
    __storm_primary__ = 'subteam_name', 'superteam_name'

    subteam_name = Unicode(name="name")
    superteam_name = Unicode(name="team")


class Meta(object):

    __storm_table__ = "meta"
    __storm_primary__ = 'key', 'value', 'blueprint_name', 'date'

    key = Unicode()
    value = Unicode()
    blueprint_name = Unicode(name="spec")
    blueprint = Reference(blueprint_name, Blueprint.name)
    date = Date(default_factory=current_date)


class Complexity(object):

    __storm_table__ = "complexity"
    __storm_primary__ = (
        'assignee_name', 'milestone_name', 'blueprint_name', 'date')

    assignee_name = Unicode(name="assignee")
    points = Unicode()
    milestone_name = Unicode(name="milestone")
    milestone = Reference(milestone_name, Milestone.name)
    blueprint_name = Unicode(name="spec")
    blueprint = Reference(blueprint_name, Blueprint.name)
    date = Date(default_factory=current_date)


class Workitem(object):

    __storm_table__ = "work_items"
    __storm_primary__ = (
        'description', 'blueprint_name', 'status', 'assignee_name',
        'milestone_name', 'date')

    STATUS_TODO = u'todo'
    STATUS_DONE = u'done'
    STATUS_POSTPONED = u'postponed'
    STATUS_INPROGRESS = u'inprogress'
    STATUS_BLOCKED = u'blocked'

    VALID_STATUSES = (
        STATUS_TODO,
        STATUS_DONE,
        STATUS_POSTPONED,
        STATUS_INPROGRESS,
        STATUS_BLOCKED,
        )

    description = Unicode()
    blueprint_name = Unicode(name="spec")
    blueprint = Reference(blueprint_name, Blueprint.name)
    status = Unicode()
    assignee_name = Unicode(name="assignee")
    milestone_name = Unicode(name="milestone")
    milestone = Reference(milestone_name, Milestone.name)
    date = Date(default_factory=current_date)


class BlueprintInBlueprintGroup(object):

    __storm_table__ = "spec_group_specs"
    __storm_primary__ = "blueprint_name", "blueprint_group_name"

    blueprint_group_name = Unicode(name="spec_group")
    blueprint_name = Unicode(name="spec")


class BlueprintGroup(object):

    __storm_table__ = "spec_groups"

    name = Unicode(primary=True)
    url = Unicode()
    priority = Unicode()
    implementation = Unicode()
    assignee_name = Unicode(name="assignee")
    status = Unicode()
    milestone_name = Unicode(name="milestone")
    milestone = Reference(milestone_name, Milestone.name)
    details_url = Unicode()
    area = Unicode()
    blueprints = ReferenceSet(
        name, BlueprintInBlueprintGroup.blueprint_group_name,
        BlueprintInBlueprintGroup.blueprint_name, Blueprint.name)

    @classmethod
    def from_launchpad(cls, lp_bp):
        model_bp_group = cls()
        fill_blueprint_info_from_launchpad(model_bp_group, lp_bp)
        return model_bp_group


def extract_last_path_segment_from_url(url):
    """Extract the last path segment from the given URL.

    This is useful if you have the URL of a Launchpad API resource and want to
    get its name without going over the network. For example, if you have a
    blueprint object and want to get its milestone name you can use:

        extract_last_path_segment_from_url(bp.milestone_link)

    Where milestone_link will contain the URL resource to the blueprint's
    milestone.
    """
    return url.split('/')[-1]


def extract_user_name_from_url(url):
    """Extract a user name from the given URL.

    The URL's last path segment must be of the form ~user, and this method
    will return just the actual user name, without the tilde.
    """
    segment = extract_last_path_segment_from_url(url)
    assert segment.startswith('~'), (
        "%s doesn't seem to contain a user name" % url)
    return segment[1:]
